"""
Definitions of daemon-thread types: perpetual threads which are meant
to run for as long as the process is alive.
"""

from .thread import Thread, _ThreadStop


################################################################################

class DaemonThread(Thread):
    """
    An abstract daemon thread class.

    A daemon thread should never stop voluntarily, only when it is requested to stop
    (when the ``stop`` method is called).

    A concrete daemon thread is a subclass of ``DaemonThread``, which typically overrides
    ``_main_iteration``.  It is also possible to override ``_main`` directly (instead of
    ``_main_iteration``), in order to define a custom thread main-loop.

    Using the ``target`` argument is not supported.  The function to run in the daemon
    thread is defined by overriding the ``_main`` method.
    """

    def __init__(self, *, daemon=True, **kwargs):
        super().__init__(
            daemon=daemon,
            target=None, args=(), kwargs={},  # caller must not pass these, raises if passed
            **kwargs
        )
        self._is_premature_exit = False

    ################################################################################
    # abstract daemon implementation methods

    def _main(self):
        """
        The main logic for running the daemon thread.

        As daemon threads are supposed to be perpetual, this method should not exit prematurely,
        i.e. should not return or raise an exception before ``stop`` is called.
        """
        self._main_init()
        try:
            while not self.is_stopping():
                try:
                    self._main_iteration()
                except _ThreadStop:
                    raise
                except Exception as e:
                    self._on_error(e)
        finally:
            self._main_destroy()

    def _main_iteration(self):
        """
        A
        """
        raise NotImplementedError('_main_iteration() not defined for %s' % self.__class__.__name__)

    def stop(self, reason=None):
        """
        Signal the thread should stop executing.
        This method returns immediately. A well-behaved daemon thread exits shortly after the call.

        If this is called before the thread is started, it will exit immediately once started.
        """
        self._request_stop(reason=reason)

    ################################################################################
    # hooks

    def _on_error(self, e):
        """
        A hook for handling an exception raised from ``_main_iteration``.
        Concrete ``DaemonThread`` subclasses can override this method for customized
        error-handling.

        If an exception is raised from this method, the thread will abort.
        """
        self.logger.exception('error', exc_info=e)

    def _on_abort(self, e):
        self.logger.exception('aborted due to an error', exc_info=e)

    def _on_exit(self):
        # A daemon should only exit if requested
        if self.is_stopping():
            return  # ok
        self._is_premature_exit = True
        self._on_premature_exit()
        super()._on_exit()

    def _on_premature_exit(self):
        """
        Daemon threads should not exit prematurely.
        This method serves as a error-handler hook for handling a premature-exit condition.

        When this method is called, ``self.exception`` is set to the exception which caused the
        thread to abort, or if ``self.exception`` is None, it means ``_main`` simply returned
        (prematurely) with no error.

        Concrete ``DaemonThread`` subclasses can override this method for customized
        error-handling.
        """
        if self._stop_reason is None:
            self._stop_reason = 'premature'
        if self.exception is not None:
            self.logger.error('exiting prematurely due to an error')
        else:
            self.logger.error('exiting prematurely')

    def _main_init(self):
        pass

    def _main_destroy(self):
        pass

    ################################################################################
    # other

    def is_stopped_prematurely(self):
        return self._is_premature_exit


class EventLoopThread(DaemonThread):
    """
    A specialized DaemonThread_, customized for the common case of running an event-loop.

    The ``_main_iteration`` method is broken down to three operations:

    - ``_read_next_event`` -- abstract
    - ``_handle_event`` -- abstract
    - ``_on_event_error`` -- can be customized by the subclass

    """

    ################################################################################
    # abstract and customizable methods

    def _read_next_event(self):
        """
        Generate the next event to handle. This method is blocking.

        *Event* is an abstract notion, it can be any object at all (except ``None``).
        The event returned is passed to ``_handle_event``.

        Returning None serves for "yielding" control back to the main-loop.  It then checks
        if the thread should stop, and if not, calls this method again.

        In a well-behaved thread, this method will not block for to long, and will "yield"
        control back to the main loop (by returning None) after a short while.

        :note: This method should not return immediately with None (repeatedly), because that
        will result with a busy-wait loop.

        If this method raises an exception, ``_on_error`` is called to handle it.
        """
        raise NotImplementedError('_read_next_event() not defined for %s' %
                                  self.__class__.__name__)

    def _handle_event(self, event):
        """
        Handle the event.

        If this method raises an exception, ``_on_event_error`` is called to handle it.

        :param event: the event object returned from the last call to ``_read_next_event``.
        """
        raise NotImplementedError('_handle_event() not defined for %s' % self.__class__.__name__)

    def _on_event_error(self, event, e):
        """
        Called when ``_handle_event`` raises an exception.

        This method may be overridden for customized error handling.

        If this method raises an exception, ``_on_error`` is called to handle it.
        """
        self.logger.exception('error handling event: %s', event, exc_info=e)

    ################################################################################
    # abstract event loop implementation (private)

    def _main_iteration(self):
        # read an event:
        event = self._read_next_event()
        if event is None:
            return
        # handle the event:
        try:
            self._handle_event(event)
        except _ThreadStop:
            raise
        except Exception as e:
            self._on_event_error(event, e)


################################################################################



"""
Definitions of task-thread types: temporary threads which are meant
to perform a single task.
"""

import datetime
from concurrent.futures import CancelledError
from .thread import Thread, ThreadStatus, _ThreadStop


################################################################################

class TaskThread(Thread):
    """
    A thread which is meant to be temporary: perform a certain task and finish.

    The task can be cancelled by calling the ``cancel`` method.  When a ``TaskThread`` is
    cancelled successfully, it aborts with a `CancelledError`_.

    When the task completes, the ``result`` attribute is set with the value returned.

    :note:
        in order to be well behaved, the task should call ``_stop_if_requested`` or ``_sleep``
        often enough.
    """

    def cancel(self, reason=None):
        """
        Signal the task should be cancelled, and the thread should abort execution.

        This method returns immediately. A well-behaved daemon thread aborts shortly after
        the call, with a `CancelledError`_ (unless already done).

        If this is called before the thread is started, it will abort immediately once started
        (with a `CancelledError`_).
        """
        if self.is_stopped() or self._stopping_event.is_set():
            return
        if reason is None:
            reason = 'cancelled'
        self._request_stop(reason=reason)

    def is_cancelled(self):
        """
        Has this thread stopped due cancelling?
        """
        return self.is_stopped() and self._stopping_event.is_set()

    def _on_abort(self, e):
        if isinstance(e, CancelledError):
            self.logger.info('task cancelled')
        else:
            self.logger.info('aborted due to an error: %s', e)

    def _on_thread_stop(self, e):
        super()._on_thread_stop(e)
        raise CancelledError() from e  # raise to invoke thread-abort logic

    def _handle_stop_before_start(self):
        super()._handle_stop_before_start()
        raise _ThreadStop()  # raise to invoke self._on_thread_stop

    def status(self):
        """ A string representing the current state of the thread. """
        if self.is_cancelled():
            return ThreadStatus.cancelled
        return super().status()

    def _status_repr(self):
        # return "cancelled" instead of "cancelled (cancelled)"
        status = self.status().value
        if status == str(self._stop_reason):
            return status
        return super()._status_repr()

    def reraise(self, suppress_cancelled=False):
        """
        If the thread aborted with an error, raise it in this current (caller) thread.
        :param suppress_cancelled: if thread aborted due to cancelling, will not raise.
        """
        try:
            super().reraise()
        except CancelledError:
            if not suppress_cancelled:
                raise


class _ExpiringTaskThread(TaskThread):
    """
    An abstract thread to run a task with a predefined expiry.
    When expires, the thread exits gracefully.

    ``expiry`` can be:

     - int/float: number of seconds since thread is started
     - datetime.timedelta: time since thread is started
     - datetime.datetime: absolute expiration time

    """

    class _Expired(_ThreadStop):
        """ An internal error raised to signal it has expired and should stop executing. """
        pass

    def __init__(self, *, expiry, **kwargs):
        super().__init__(**kwargs)

        self._expiry_raw = expiry
        self._expiry = None
        self._expired = False

        # If we got an invalid expiry, report it early:
        self._calc_expiry(self._expiry_raw)

    def _on_enter(self):
        # setting expiry when starting

        super()._on_enter()
        # set expiry:
        self._expiry = self._calc_expiry(self._expiry_raw)
        # check if already expired:
        self._check_expiry()

    def _sleep(self, timeout):
        # enforcing expiry when sleeping

        # check if already expired:
        self._check_expiry()
        # modify timeout so we don't sleep beyond expiry:
        expire_after_sleep = False
        max_timeout = (self._expiry - self._now()).total_seconds()
        if timeout > max_timeout:
            timeout = max(0, max_timeout)
            expire_after_sleep = True
        # sleep:
        super()._sleep(timeout)
        # check if expired after sleeping:
        if expire_after_sleep:
            raise self._Expired()
        self._check_expiry()  # just in case

    def _on_thread_stop(self, e):
        # handle exit-on-expiry

        if isinstance(e, self._Expired):
            # expired -- not an error condition, so suppress error
            self._expired = True
            return self._on_expiry()
        return super()._on_thread_stop(e)

    def _check_expiry(self):
        if self._now() >= self._expiry:
            raise self._Expired()

    def _calc_expiry(self, expiry_raw):
        if isinstance(expiry_raw, datetime.datetime):
            # this is already the expiration time
            return expiry_raw
        elif isinstance(expiry_raw, datetime.timedelta):
            # delta relative to now
            return self._now() + expiry_raw
        elif isinstance(expiry_raw, (int, float)):
            # number of seconds, relative to now
            return self._now() + datetime.timedelta(seconds=expiry_raw)
        else:
            raise TypeError('Invalid expiry: %r' % expiry_raw)

    def is_expired(self):
        return self._expired

    def _on_expiry(self):
        """
        A hook which controls what to do when the thread expires.
        The value returned will be used as task's result.
        """
        raise NotImplementedError


class LimitedTimeTaskThread(_ExpiringTaskThread):
    """
    A `TaskThread` which runs for a predefined period of time, and then finishes
    successfully.
    """

    def _on_expiry(self):
        # finish successfully upon expiry:
        return None


class TimeoutTaskThread(_ExpiringTaskThread):
    """
    A `TaskThread` which runs for no longer than a given timeout.
    When it times out, it aborts with `TimeoutError` exception.
    """

    def _on_expiry(self):
        # abort with a TimeoutError upon expiry:
        raise TimeoutError()

    def is_timed_out(self):
        return self.is_expired()


################################################################################

class FunctionThread(TaskThread):
    """
    A thread for running a given function, passed as the ``target`` argument.

    This class is provided for convenience.  It is not a well-behaved thread,
    as it doesn't support cancelling the task (it only supports cancelling before the
    thread is started).

    It is recommended to use `TaskThread`_ instead, i.e. defining a subclass of
    ``TaskThread`` which performs the task well-behaved-ly.
    """

    def __init__(self, target, *, name=None, **kwargs):
        if name is None:
            try:
                name = target.__name__
            except Exception:
                name = str(target)
        super().__init__(target=target, name=name, **kwargs)

    def _main(self):
        try:
            return self._target(*self._args, **self._kwargs)
        finally:
            # Avoid a refcycle if the thread is running a function with
            # an argument that has a member that points to the thread.
            del self._target, self._args, self._kwargs

    def cancel(self, reason=None):
        if self.is_alive():
            raise RuntimeError('FunctionThread is already running and cannot be cancelled')
        return super().cancel(reason=reason)


################################################################################



"""
Definitions of *merethread* ``Thread`` baseclass.
"""

import datetime
from enum import Enum
import threading as threading
from concurrent.futures import Future
import lo99ing

from .misc import Runtime, ProfileContext, NoopContext, get_currnet_stacktrace


################################################################################
# Misc classes

class _ThreadStop(Exception):
    """ An internal error raised in a thread to signal it should stop executing. """
    pass


class ThreadStatus(Enum):
    not_started = 'no started'
    running = 'running'
    stopping = 'stopping'
    stopped = 'stopped'
    aborted = 'aborted'
    cancelled = 'cancelled'
    stopped_before_starting = 'stopped before starting'


class ThreadFuture(Future):
    """
    A Future_, with minor improvements and adjustments for making it slightly more suitable
    for "encapsulating the asynchronous execution" of a *thread*.

    Notable additions: the ``thread`` attribute, and the ``add_callback``/``add_errback`` methods.
    """

    def __init__(self, thread, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.thread = thread

    def add_callback(self, fn):
        """
        Same as ``add_done_callback``, but the only invoked in case of success.

        ``fn`` is passed two positional arguments: the thread and the result.
        """

        def cb(fut):
            try:
                result = fut.get_result()
            except Exception:
                # error, so not calling the callback
                pass
            else:
                return fn(fut.thread, result)

        self.add_done_callback(cb)

    def add_errback(self, fn):
        """
        Same as ``add_done_callback``, but only invoked in case of an error.

        If the future was cancelled, ``fn`` will be called with the ``CancelledError`` exception.

        ``fn`` is passed two positional arguments: the thread and the exception.
        """

        def eb(fut):
            try:
                fut.get_result()
                # no error, so not calling the errback
            except Exception as e:
                return fn(fut.thread, e)

        self.add_done_callback(eb)

    def cancel(self):
        raise NotImplementedError(
            'Cancelling a thread is not supported. '
            'Use DaemonThread.stop() or TaskThread.cancel() instead '
            '(e.g. future.thread.cancel()).')


################################################################################
# The MereThread thread baseclass

class Thread(threading.Thread):
    """
    The *merethread* ``Thread`` baseclass.
    This class is merely a subclass of `threading.Thread`_, which adds various useful features and
    capabilities to it.

    In most cases, you wouldn't want to use this class directly, but a subclass
    suitable for your use case, e.g ``DaemonThread``, ``TaskThread``, etc.

    This class adds a few features to threads (see the docs for details):

    - ``exception`` attribute for accessing the exception which caused thread to abort.
    - ``_result`` attribute for accessing the value returned by the ``_main`` method.
    - A `concurrent.futures.Future`_ interface, using the ``future`` attribute.
    - Clean stopping of the thread: this is implemented in this base class, but exposed in
        subclasses. See ``DaemonThread`` and ``TaskThread`` for example.
    - Easily run with a profiler.
    - Easily view the current (live) stack-trace of the thread.
    - ``runtime`` attribute for accessing thread execution start/end times.
    - The ``join`` method returns a bool indicating whether thread has finished.

    :note:
        Unlike usage of the `threading.Thread`_ class, when subclassing ``merethread.Thread`` you
        DO NOT override the ``run`` method.  You should override the ``_main`` method instead.
        This class defines ``run``, and the entry point for custom subclasses is the ``_main``
        method.

    To support stopping, the ``_main`` method should call the ``_stop_if_requested`` or ``_sleep``
    methods often. See the *Well Behaved Threads* section of the docs for more details.
    """

    # Implementation note: to avoid problems related to issue #18808, we avoid accessing
    # the _is_stopped member directly.  Calling is_alive() is a safer option, and is appropriate
    # in most cases.

    Future = ThreadFuture
    Runtime = Runtime
    ProfileContext = ProfileContext

    ################################################################################
    # constructor

    def __init__(self, *,
                 logger=None, logger_name=None, clock=None,
                 profile=False, profile_kwargs=None,
                 **kwargs):
        """
        :param profile: If True, the thread will run with profiling enabled, using ProfileContext_.
        :param profile_kwargs: extra kwargs to pass to the ``ProfileContext``.
        """

        super().__init__(**kwargs)

        if logger is None:
            if logger_name is None:
                logger_name = '%s' % self.name
            logger = lo99ing.get_logger(logger_name)
        self.logger = logger
        if clock is None:
            clock = datetime.datetime.now
        self._clock = clock

        self._stopping_event = threading.Event()
        self._stop_reason = None

        if profile_kwargs is None:
            profile_kwargs = {}
        self._profiler_ctx = self._get_profiler_ctxmgr(profile, **profile_kwargs)

        # The following are private. subclasses should not set them:
        self.__result = None
        self.__exception = None
        self.__runtime = self.Runtime(clock=self._clock)
        self.__future = self.Future(self)

    ################################################################################
    # main

    def run(self):
        """
        The basic *merethread* main logic.

        DO NOT OVERRIDE THIS METHOD.  You should override ``_main`` instead.
        """

        with self._profiler_ctx, self.__runtime:

            result = None
            exception = None

            try:

                self.future.set_running_or_notify_cancel()

                # pre-start checks
                if not self._stopping_event.is_set():
                    # starting
                    self._on_enter()
                    # main
                    result = self._main()
                else:
                    # thread stop requested before we got a chance to start running
                    self._handle_stop_before_start()

                return result

            except _ThreadStop as e:
                # stop has been requested
                try:
                    result = self._on_thread_stop(e)  # may raise or not
                except Exception as e2:
                    exception = e2

            except Exception as e:
                exception = e

            finally:

                # set self.__result and self.__exception
                if exception is not None:
                    self.__exception = exception
                    self._on_abort(self.__exception)
                else:
                    self.__result = result

                # set self.future with the result/exception:
                if not self.future.cancelled():
                    if exception is not None:
                        self.future.set_exception(exception)
                    else:
                        self.future.set_result(result)

                # other cleanups:
                self._on_exit()

    def _main(self):
        """
        Method representing the thread's activity.

        Subclasses should override the ``_main`` method instead of ``run``.
        """
        return super().run()

    ################################################################################
    # stopping

    def _request_stop(self, reason=None):
        """
        Signal the thread it should stop as soon as possible.

        A well-behaved thread will stop shortly after ``_request_stop`` is called.

        Can be called from any thread.
        """
        self.logger.info('stop requested (%s)', reason)
        self._stop_reason = reason
        self._stopping_event.set()

    def _stop_if_requested(self):
        """
        Raises `_ThreadStop`_ if ``_request_stop`` has already been called.
        """
        self._sleep(0)

    def _sleep(self, timeout):
        """
        Sleep for ``timeout`` seconds, or until ``_request_stop`` is called.

        :raise _ThreadStop: if ``_request_stop`` is called while (or prior to) sleeping.
        """
        is_stopping = self._stopping_event.wait(timeout)
        if is_stopping:
            raise _ThreadStop()

    ################################################################################
    # state

    def is_started(self):
        """
        Has this thread been started (whether finished or not)?
        """
        return self._started.is_set()

    def is_stopping(self):
        """
        Is this thread being stopped?

        A thread is in a "stopping" state if self._request_stop() has been called, but the thread
        is still alive.
        """
        return self._stopping_event.is_set() and self.is_alive()

    def is_stopped(self):
        """
        Has this thread already stopped?
        """
        return self.is_started() and not self.is_alive()

    def is_aborted(self):
        """
        Has this thread stopped due to an error?
        """
        return self.__exception is not None and not self.is_alive()

    def status(self):
        """
        A `ThreadStatus`_ representing the current status of the thread.
        """
        stopping = self._stopping_event.is_set()
        if not self.is_started():
            return ThreadStatus.stopped_before_starting if stopping else ThreadStatus.not_started
        elif self.is_alive():
            return ThreadStatus.stopping if stopping else ThreadStatus.running
        elif self.is_aborted():
            return ThreadStatus.aborted
        else:
            assert self.is_stopped()
            return ThreadStatus.stopped

    ################################################################################
    # hooks

    def _on_enter(self):
        """
        A hook which is called when the thread starts executing.

        Will not be called if the thread is requested to stop before it started.

        If an exception is raised, the thread will abort with that exception.
        """
        self.logger.info('starting')

    def _on_exit(self):
        """
        A hook which is called as the last thing to do before the thread finishes.

        Called regardless of whether the thread finished successfully or aborted due to
        an error.
        """
        self.logger.info('done')

    def _on_abort(self, e):
        """
        A hook for handling (reporting) the case where the thread is aborting due to
        an exception.

        The ``_on_exit`` hook will be called after this.

        No exception should be raised from this method.
        """
        pass

    def _on_thread_stop(self, e):
        """
        A hook which controls what to do when the thread detects it has been requested to stop.

        If an exception is raised, the thread will abort with that exception.
        Else, the thread will exit cleanly, and result will be set to the value returned.
        """
        self.logger.info('stopping')

    def _handle_stop_before_start(self):
        """
        A hook which controls what to do when the thread detects it has been requested to stop
        before it started to run.

        If this is called, ``_on_enter`` and ``_main`` are not called.

        If an exception is raised, the thread will abort with that exception.
        Else, the thread will exit cleanly.
        """
        self.logger.info('not starting, stop already requested')

    ################################################################################
    # profiling, debugging, introspection

    def _get_profiler_ctxmgr(self, profile, **kwargs):
        if profile:
            return self.ProfileContext(**kwargs)
        else:
            return NoopContext()

    @property
    def profiler(self):
        """
        A profiler object which can be used for accessing profiler stats for this thread.

        This is None if profiling has not been enabled (using the ``profile`` flag)
        """
        return getattr(self._profiler_ctx, 'profiler', None)

    @property
    def runtime(self):
        """
        A Runtime_ object capturing thread's start/end times.
        """
        return self.__runtime

    def get_current_stacktrace(self):
        """
        :return: a multiline string capturing the current stack-trace of this thread.
        """
        return get_currnet_stacktrace(self)

    ################################################################################
    # other

    @property
    def future(self):
        """
        A ThreadFuture_ which can be used for adding callbacks (and errbacks) to be
        called when the thread finishes.

        Useful mainly for ``TaskThread``s.

        :note: The thread *cannot* be cancelled using ``t.future.cancel()``.
        """
        return self.__future

    @property
    def result(self):
        """
        If the thread finished with no error, this is the value returned from its ``_main``.
        Else (hasn't finished yet, or aborted), this is None.
        """
        return self.__result

    @property
    def exception(self):
        """
        If the thread aborted, this is the exception which caused it to abort.
        Else (hasn't finished yet, or finished with no error), this is None.
        """
        return self.__exception

    def reraise(self):
        """
        If the thread aborted with an error, raise it in this current (caller) thread.
        """
        if self.__exception is not None:
            raise self.__exception

    def join(self, timeout=None):
        """
        Same as `threading.Thread.join`_, but also returns a flag indicating whether
        the operation completed with no timeout (similar to ``Event.wait``, ``Lock.acquire``,
        etc).

        From the documentation of the baseclass::

            As join() always returns None, you must call isAlive() after join() to decide whether
            a timeout happened -- if the thread is still alive, the join() call timed out.

        This method does this for you, so you don't have to call is_alive().

        :return: False iff returned due to a timeout.
        """
        super().join(timeout)
        return not self.is_alive()

    def _now(self):
        return self._clock()

    def __repr__(self):
        assert self._initialized, "Thread.__init__() was not called"

        status = self._status_repr()
        if self.is_stopped():
            status += ', %s' % self.runtime
        status = '[%s]' % status

        return '<%s>' % ' '.join(
            str(x) for x in
            [self.__class__.__name__, self.name, self.ident, status]
            if x
        )

    def _status_repr(self):
        status = self.status().value
        if self._stop_reason is not None:
            status += ' (%s)' % self._stop_reason
        return status


################################################################################



"""
Miscellaneous tools used in this package.
"""

import sys
import traceback
import datetime
import cProfile


################################################################################

class Runtime:
    """
    A context-manager which records block-execution start-time and end-time.
    """

    def __init__(self, clock=None):
        if clock is None:
            clock = datetime.datetime.now
        self.clock = clock
        self.start = None
        self.end = None

    def set_start(self, t=None):
        if t is None:
            t = self.now()
        self.start = t
        self.end = None

    def set_end(self, t=None):
        if not self.is_started:
            raise RuntimeError('not started')
        if t is None:
            t = self.now()
        self.end = t

    @property
    def is_started(self):
        return self.start is not None

    @property
    def is_ended(self):
        return self.end is not None

    @property
    def total(self):
        """ end minus start, as timedetla """
        if self.is_ended:
            return self.end - self.start

    @property
    def total_seconds(self):
        """ same as ``total``, but in seconds """
        dt = self.total
        if dt is not None:
            return dt.total_seconds()

    def now(self):
        return self.clock()

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self)

    def __str__(self):
        if not self.is_started:
            return 'not started'
        if not self.is_ended:
            return 'running'
        return 'finished after %s' % self.total

    ################################################################################
    # context manager

    def __enter__(self):
        self.set_start()
        self.end = None
        return self

    def __exit__(self, type, value, tb):
        self.set_end()


################################################################################

def get_currnet_stacktrace(thread):
    """
    Returns a multiline string capturing the current stack-trace of the given thread.

    Based on: https://stackoverflow.com/a/2569696
    """
    cur_frame = sys._current_frames().get(thread.ident)
    if cur_frame is None:
        return None
    code = []
    code.append('\n# Thread: %s(%d)' % (thread.name, thread.ident))
    for filename, lineno, name, line in traceback.extract_stack(cur_frame):
        code.append('File: "%s", line %d, in %s' % (filename, lineno, name))
        if line:
            code.append('  %s' % (line.strip()))
    return '\n'.join(code)


################################################################################

class ProfileContext:
    """
    A context manager for profiling blocks.
    """

    def __init__(self, profiler=None,
                 print_stats=False, print_stats_kwargs=None,
                 dump_to_file=None):
        """
        :param print_stats: if True, will print stats on exit
        :param dump_to_file: if passed, will ``dump_stats`` to this path on exit
        """
        if profiler is None:
            profiler = cProfile.Profile()
        self.profiler = profiler
        self.print_stats = print_stats
        self.print_stats_kwargs = print_stats_kwargs
        self.dump_to_file = dump_to_file

    def __enter__(self):
        self.profiler.enable()

    def __exit__(self, type, value, tb):
        self.profiler.disable()

        if self.print_stats:
            kwargs = self.print_stats_kwargs or {}
            try:
                self.profiler.print_stats(**kwargs)
            except Exception:
                # __exit__ must not raise
                pass

        if self.dump_to_file is not None:
            try:
                self.profiler.dump_stats(self.dump_to_file)
            except Exception:
                # __exit__ must not raise
                pass


################################################################################

class NoopContext:
    """ A context manager which does nothing. """

    def __enter__(self):
        pass

    def __exit__(self, *args, **kwargs):
        return False


################################################################################



"""
The ``merethread`` package, providing thread classes with various useful features, as well as
customized ``threading.Thread`` subclasses suitable for the most common cases of using threads.

For details, see the docs.
"""

from .thread import Thread, ThreadStatus
from .daemon import DaemonThread, EventLoopThread
from .task import TaskThread, FunctionThread

Thread, ThreadStatus, DaemonThread, EventLoopThread, TaskThread, FunctionThread  # pyflakes



"""
Concrete thread types, which are not useful in general.
Meant mainly for demonstrating how to define a variety of thread types.
Some of these are also used for unittesting.
"""

import time
from .thread import Thread
from .daemon import DaemonThread, EventLoopThread
from .task import TaskThread, FunctionThread, LimitedTimeTaskThread, TimeoutTaskThread


################################################################################

SAMPLE_RESULT = 42
SAMPLE_EXCEPTION = RuntimeError('failed intentionally')


################################################################################
# Thread samples

class IdleThread(Thread):
    """ A thread which sleeps indefinitely """

    def _main(self):
        self._sleep()

    def stop(self, reason=None):
        return self._request_stop(reason=reason)


class IdleThreadTARGET(Thread):
    """ Same as `IdleThread`_, but using the ``target`` argument. """

    def __init__(self, **kwargs):
        super().__init__(target=_idle, args=(self,), **kwargs)

    def stop(self, reason=None):
        return self._request_stop(reason=reason)


class NoopThread(Thread):
    """ A thread which does nothing and returns immediately """
    def _main(self):
        return


class NoopThreadTARGET(Thread):
    """ Same as `NoopThread`_, but using the ``target`` argument. """
    def __init__(self, **kwargs):
        super().__init__(target=_noop, **kwargs)


class AbortingThread(Thread):
    """ A thread which aborts immediately """

    EXCEPTION_TYPE = type(SAMPLE_EXCEPTION)

    def _main(self):
        _fail()


class AbortingThreadTARGET(Thread):
    """ Same as `AbortingThread`_, but using the ``target`` argument. """

    EXCEPTION_TYPE = type(SAMPLE_EXCEPTION)

    def __init__(self, **kwargs):
        super().__init__(target=_fail, **kwargs)


################################################################################
# DaemonThread samples

class IdleDaemonThread(DaemonThread):
    """ A daemon thread which sleeps indefinitely """
    def _main_iteration(self):
        self._sleep()


class ReturningDaemonThread(DaemonThread):
    """ A daemon which returns immediately (not a valid behavior for a daemon) """
    def _main(self):
        return


class AbortingDaemonThread(DaemonThread):
    """ A daemon which aborts immediately (not a valid behavior for a daemon) """

    EXCEPTION_TYPE = type(SAMPLE_EXCEPTION)

    def _main(self):
        _fail()


class MetronomeDaemonThread(DaemonThread):
    """ A metronome, printing tick and tock """

    def __init__(self, period=1, **kwargs):
        super().__init__(**kwargs)
        self.period = period
        self.count = 0

    def _main_iteration(self):
        self._sleep(self.period / 2.)
        msg = ['Tick', 'tocK'][self.count % 2]
        self.count += 1
        print(msg)


class MetronomeEventLoopThread(EventLoopThread):
    """ An event-loop based metronome """

    def __init__(self, period=1, **kwargs):
        super().__init__(**kwargs)
        self.period = period
        self.count = 0

    def _read_next_event(self):
        self._sleep(self.period / 2.)
        event = ['Tick', 'tocK'][self.count % 2]
        self.count += 1
        return event

    def _handle_event(self, event):
        print(event)


class FaultyMetronomeEventLoopThread(MetronomeEventLoopThread):
    """ An event-loop based metronome, which sometime fails handling an event """

    def _handle_event(self, event):
        if self.count % 10 == 0:
            raise RuntimeError('glitch')
        return super()._handle_event(event)


################################################################################
# TaskThread samples

class NoopTaskThread(TaskThread):
    """ A task which returns immediately """

    RESULT = SAMPLE_RESULT

    def _main(self):
        return self.RESULT


class IdleTaskThread(TaskThread):
    """ A task which sleeps for a predefined period """

    RESULT = SAMPLE_RESULT

    def __init__(self, period=10, **kwargs):
        super().__init__(**kwargs)
        self.period = period

    def _main(self):
        self._sleep(self.period)
        return self.RESULT


class FailedTaskThread(TaskThread):
    """ A task which raises an exception immediately """

    EXCEPTION_TYPE = type(SAMPLE_EXCEPTION)

    def _main(self):
        _fail()


class SlowTaskThread(TaskThread):
    """ Find a prime number greater than N, very inefficiently """

    def __init__(self, min=20000000, **kwargs):
        super().__init__(**kwargs)
        self.min = min

    def _main(self):
        x = self.min
        while True:
            for d in range(2, x):
                self._stop_if_requested()
                if x % d == 0:
                    break
            else:
                return x
            x += 1


def noop_function_thread():
    """ Same as `NoopTaskThread`_, but implemeted using a `FunctionThread`_ """
    return FunctionThread(_noop_func)


def idle_function_thread(*args, **kwargs):
    """ Same as `IdleTaskThread`_, but implemeted using a `FunctionThread`_ """
    return FunctionThread(_long_func, args=args, kwargs=kwargs)


def failed_function_thread():
    """ Same as `FailedTaskThread`_, but implemeted using a `FunctionThread`_ """
    return FunctionThread(_fail)


################################################################################
# ExpiringTaskThread samples

class NoopLimitedTimeTaskThread(LimitedTimeTaskThread):
    """ A task which returns immediately """

    RESULT = SAMPLE_RESULT

    def _main(self):
        return self.RESULT


class IdleLimitedTimeTaskThread(LimitedTimeTaskThread):
    """ A task which sleeps until expired """

    def _main(self):
        self._sleep(9999999999)


class FailedLimitedTimeTaskThread(LimitedTimeTaskThread):
    """ A task which raises an exception immediately """

    EXCEPTION_TYPE = type(SAMPLE_EXCEPTION)

    def _main(self):
        _fail()


class NoopTimeoutTaskThread(TimeoutTaskThread):
    """ A task which returns immediately """

    RESULT = SAMPLE_RESULT

    def _main(self):
        return self.RESULT


class IdleTimeoutTaskThread(TimeoutTaskThread):
    """ A task which sleeps until expired """

    def _main(self):
        self._sleep(9999999999)


class FailedTimeoutTaskThread(TimeoutTaskThread):
    """ A task which raises an exception immediately """

    EXCEPTION_TYPE = type(SAMPLE_EXCEPTION)

    def _main(self):
        _fail()


################################################################################
# misc

def _noop():
    return


def _noop_func():
    return SAMPLE_RESULT


def _long_func(period=10):
    time.sleep(period)
    return SAMPLE_RESULT


def _fail():
    raise SAMPLE_EXCEPTION


def _idle(thread):
    thread._sleep()


################################################################################



"""
Potentially useful tools for working with threads.
"""

from . import TaskThread


################################################################################

class ThreadLifeCycleContext:
    """
    A context manager for running a thread (or threads) for as long as the enclosed block runs,
    i.e. starting it when the block starts, and stopping it when it is done.

    Useful for various types of monitoring of block execution, e.g. the thread can periodically
    print memory-consumption while a (long-running) block runs.

    On `__enter__`, the thread is started (if haven't been started before).
    On `__exit__`, the thread is (optionally) stopped, and (optionally) joined.

    To be clear, the threads are oblivious to the code running in the block.
    They run their own code, and the code in the block runs in the original caller thread.
    """

    def __init__(self,
                 *threads,
                 stop=True, join=True, join_timeout=None, raise_on_timeout=True,
                 reraise=True,
                 suppress_keyboard_interrupt=True):
        """
        :param threads: `merethread.Thread`s
        :param stop: whether to stop threads on __exit__
        :param join: whether to join threads on __exit__
        :param join_timeout: timeout to pass to `thread.join()`, if `join=True`
        :param raise_on_timeout: whether to raise a `TimeoutError` if `thread.join()` times out
        :param reraise: whether to reraise an exception causing one of the threads to abort
        :param suppress_keyboard_interrupt: whether to propagate KeyboardInterrupt exception
        """
        self.threads = threads
        self.should_stop = stop
        self.should_join = join
        self.join_timeout = join_timeout
        self.should_raise_on_timeout = raise_on_timeout
        self.should_reraise = reraise
        self.suppress_keyboard_interrupt = suppress_keyboard_interrupt
        self._stopped = False

    def __enter__(self):
        for thread in self.threads:
            if not thread.is_started():
                thread.start()

    def __exit__(self, exc_type, exc_value, tb):

        # note: we always stop in case of KeyboardInterrupt, even if self.should_stop=False
        if self.should_stop or isinstance(exc_value, KeyboardInterrupt):
            self._stop()

        new_exception = None

        if self.should_join:
            interrupt = None
            for thread in reversed(self.threads):
                try:
                    self._join_thread(thread)
                except TimeoutError as e:
                    new_exception = e
                    break
                except KeyboardInterrupt as e:
                    interrupt = e
                    new_exception = e
                    break
                except Exception as e:
                    new_exception = e
                    break

            # handle KeyboardInterrupt raised while joining:
            if interrupt is not None:

                if not self._stopped:
                    # note: we stop in case of KeyboardInterrupt, even if self.should_stop=False
                    self._stop()
                    # now join again, after stopping
                    for thread in reversed(self.threads):
                        try:
                            self._join_thread(thread)
                        except TimeoutError as e:
                            return self._report_exception(exc_value, e)
                        except Exception as e:
                            return self._report_exception(exc_value, e)

                return self._report_exception(exc_value, interrupt)

        return self._report_exception(exc_value, new_exception)

    def _stop(self):
        self._stopped = True
        for thread in reversed(self.threads):
            self._stop_thread(thread)

    def _stop_thread(self, thread):
        if not thread.is_alive():
            return
        reason = '%s.__exit__' % type(self).__name__
        if isinstance(thread, TaskThread):
            thread.cancel(reason=reason)
        else:
            thread.stop(reason=reason)

    def _join_thread(self, thread):
        is_done = thread.join(self.join_timeout)
        if is_done:
            if self.should_reraise:
                self._reraise(thread)
        else:
            if self.should_raise_on_timeout:
                raise TimeoutError(thread)

    def _reraise(self, thread):
        kwargs = {}
        if isinstance(thread, TaskThread):
            kwargs.update(suppress_cancelled=True)
        thread.reraise(**kwargs)

    def _report_exception(self, original_exception, new_exception=None):
        # suppress_keyboard_interrupt if relevant:
        if self.suppress_keyboard_interrupt:
            if isinstance(original_exception, KeyboardInterrupt):
                original_exception = None
            if isinstance(new_exception, KeyboardInterrupt):
                new_exception = None
        # the original exception takes precedence over the new exception
        if original_exception is not None:
            return False  # indicate caller to raise original exception
        elif new_exception is not None:
            # report the new exception
            raise new_exception
        else:
            return True  # no exception


################################################################################
